#| -*-Scheme-*-

Copyright (C) 2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a devops plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Developer Operations

;;; See devops.texi for complete details.

;;; package: (devops build)
;;; package: (user)  when downloaded to a build host.

(define project-name)
(define build-dir)
(define build-scheme-architecture)
(define build-debian-architecture)
(define build-ubuntu?)

(define (build)
  (in-batch
   (lambda ()
     (let loop ()
       (let ((todo (find-work)))
	 (log (number->string (length todo) 10)
	      " binaries to build.\n")
	 (flush-output-port)
	 (if (null? todo)
	     'done
	     (begin
	       (dynamic-wind grab-lock-file
			     (lambda () (for-each (lambda (work) (work)) todo))
			     release-lock-file)
	       (log "Looking for more work.\n")
	       (flush-output-port)
	       (loop))))))))

(define (find-work)
  (let ((srcs (available-sources build-dir)))
    (log "Available sources:\n")
    (for-each (lambda (src)
		(let ((name (car src))
		      (vers (cdr src)))
		  (log "    "name" "vers"\n")))
	      srcs)
    (append-map!
     (lambda (name.vers)
       (let ((name (car name.vers))
	     (vers (cdr name.vers)))
	 (if (string=? project-name name)
	     (append
	      (work-todo name vers build-scheme-architecture "pkg"
			 (lambda () (build-core-pkg name vers)))
	      (if build-ubuntu?
		  (work-todo name vers build-debian-architecture "dpkg"
			     (lambda () (build-core-dpkg name vers)))
		  '()))
	     (append
	      (work-todo name vers build-scheme-architecture "pkg"
			 (lambda () (build-plugin-pkg name vers)))
	      (if build-ubuntu?
		  (work-todo name vers build-debian-architecture "dpkg"
			     (lambda () (build-plugin-dpkg name vers)))
		  '())))))
     srcs)))

(define (work-todo name vers arch type work)
  (let ((elogfile (string build-dir"/"name"-"vers"-"arch"-"type"-error.log"))
	(logfile (string build-dir"/"name"-"vers"-"arch"-"type".log")))
    (if (file-exists? elogfile)
	(error "Failed or in progress:" elogfile))
    (if (file-exists? logfile)
	'()
	(list (lambda ()
		(log-timestamp)
		(log "Building "name" "vers" "arch" "type"\n")
		(with-output-log
		 elogfile
		 (lambda ()
		   (log-timestamp)
		   (work)
		   (log-timestamp)))
		(rename-file elogfile logfile))))))

(define (build-core-pkg name vers)
  (let ((sarch build-scheme-architecture)
	(pkgdir (string build-dir"/"name"-"vers))
	(options (string
		  (if (flag-day?)
		      " --enable-cross-compiling"
		      "")
		  (if (string-prefix? "svm" build-scheme-architecture)
		      (string " --enable-native-code="build-scheme-architecture)
		      ""))))
    (run "rm -rf "pkgdir)
    (run "cd "build-dir" && tar xzf "name"-"vers".tar.gz")
    (run "cd "pkgdir"/src && ./configure" options)
    (run "cd "pkgdir"/src && make")
    (run "cd "pkgdir"/src/microcode && make distclean")
    (run "cd "pkgdir"/doc && ./configure")
    (run "cd "pkgdir"/doc && make")
    (run "chmod -R go-w "pkgdir)
    (run "cd "build-dir" && tar czf "name"-"vers"-"sarch".tar.gz "name"-"vers)
    (set-file-modes! (string pkgdir"-"sarch".tar.gz") #o444)
    (run "rm -rf "pkgdir)
    (run "cd "build-dir" && tar xzf "name"-"vers"-"sarch".tar.gz")
    (run "cd "pkgdir"/src && ./configure" options)
    (run "cd "pkgdir"/src && make")
    (run "cd "pkgdir"/src && umask 022 && make install")
    (run "cd "pkgdir"/doc && ./configure")
    (run "cd "pkgdir"/doc && umask 022 && make install install-pdf")
    (run "rm -rf "pkgdir)))

(define (build-core-dpkg name vers)
  (let ((pkgdir (string build-dir"/"name"-"vers)))
    (run "rm -rf "pkgdir)
    (run "cd "build-dir" && tar xJf "name"_"vers".tar.xz")
    (if (flag-day?)
	(with-subprocess-environment-variable
	 "CONF_FLAGS" "--enable-cross-compiling"
	 (lambda ()
	   (run "cd "pkgdir" && dpkg-buildpackage -b -uc")))
	(run "cd "pkgdir" && dpkg-buildpackage -b -uc"))
    (let ((darch build-debian-architecture))
      (set-file-modes! (string build-dir"/"name"_"vers"_"darch".deb") #o444)
      (set-file-modes! (string build-dir"/"name"_"vers"_"darch".changes") #o444)
      (let ((f (string build-dir"/"name"_"vers"_"darch".buildinfo")))
	(if (file-exists? f) (set-file-modes! f #o444)))
      (let ((f (string build-dir"/"name"-dbgsym_"vers"_"darch".ddeb")))
	(if (file-exists? f) (set-file-modes! f #o444)))
      (run "rm -rf "pkgdir)
      (run "sudo "project-name"-install "name"_"vers"_"darch".deb"))))

(define (flag-day?)
  (and (equal? '(10 1 10) (get-subsystem-version "Release"))
       (equal? 'x86-64 microcode-id/compiled-code-type)))

(define (build-plugin-pkg name vers)
  (let ((pkgdir (string build-dir"/"name"-"vers)))
    (run "rm -rf "pkgdir)
    (run "cd "build-dir" && tar xzf "name"-"vers".tar.gz")
    (with-subprocess-environment-variable
     "MIT_SCHEME_EXE" (string "/usr/local/bin/"project-name)
     (lambda ()
       (run "cd "pkgdir" && ./configure")
       (run "cd "pkgdir" && make all check")
       (run "cd "pkgdir" && umask 022"
	    " && make install install-html install-pdf")))
    (run "rm -rf "pkgdir)))

(define (build-plugin-dpkg name vers)
  (let ((pkgdir (string build-dir"/"name"-"vers)))
    (run "rm -rf "pkgdir)
    (run "cd "build-dir" && tar xJf "name"_"vers".tar.xz")
    (with-subprocess-environment-variable
     "MIT_SCHEME_EXE" (string "/usr/bin/"project-name)
     (lambda ()
       (run "cd "pkgdir" && dpkg-buildpackage -b -uc")))
    (let ((darch build-debian-architecture))
      (set-file-modes! (string build-dir"/"name"_"vers"_"darch".deb") #o444)
      (set-file-modes! (string build-dir"/"name"_"vers"_"darch".changes") #o444)
      (let ((f (string build-dir"/"name"_"vers"_"darch".buildinfo")))
	(if (file-exists? f) (set-file-modes! f #o444)))
      (let ((f (string build-dir"/"name"-dbgsym_"vers"_"darch".ddeb")))
	(if (file-exists? f) (set-file-modes! f #o444)))
      (run "rm -rf "pkgdir)
      (run "sudo "project-name"-install "name"_"vers"_"darch".deb"))))

(define (grab-lock-file)
  (let ((filename (string build-dir"/lock")))
    (log filename(if (file-exists? filename)
		     " locked\n"
		     " free\n"))
    (call-with-exclusive-output-file
     filename
     (lambda (out)
       (write (unix/current-pid) out)))
    (log "Lock grabbed.\n")))

(define (release-lock-file)
  (log-timestamp)
  (delete-file (string build-dir"/lock"))
  (log "Lock released.\n"))

;;;; Common procedures
;;;
;;; These are not dependent on the build environment (i.e. build-dir,
;;; etc.) and are shared with the rest of devops.

(define (ubuntu?)
  (and (file-exists? "/etc/lsb-release")
       (find (lambda (line) (string=? "DISTRIB_ID=Ubuntu" line))
	     (file-lines "/etc/lsb-release"))))

(define (read-debian-architecture)
  (car (shell-lines "dpkg-architecture -qDEB_TARGET_ARCH")))

(load-option 'regular-expression)

(define (available-sources dir)
  (map cdr
       (sort
	(append-map!
	 (lambda (name)
	   (let ((regs (re-string-match
			"\\(.*\\)-\\([0-9]+[.][0-9.]+\\)[.]tar[.]gz$"
			name)))
	     (if regs
		 (list (cons* (file-modification-time (string dir"/"name))
			      (re-match-extract name regs 1)
			      (re-match-extract name regs 2)))
		 '())))
	 (directory-file-names dir #f))
	(lambda (a b) (< (car a) (car b))))))

(define (in-batch thunk)
  (fresh-line)
  (write-string "OK\n")
  (flush-output-port)
  (exit
   (with-output-to-file (string build-dir"/daemon.log")
     (lambda ()
       (call-with-current-continuation
	(lambda (abort-batch-job)
	  (bind-condition-handler '()
	      (named-lambda (batch-condition-handler condition)
		(fresh-line)
		(log-timestamp)
		(write-condition-report condition (current-output-port))
		(newline)
		(flush-output-port)
		(abort-batch-job 7))
	    (lambda ()
	      (log-timestamp)
	      (flush-output-port)
	      (log "Batch job started.\n")
	      (flush-output-port)
	      (thunk)
	      (log "Batch job succeeded.\n")
	      0))))))))

(define (with-output-log filename thunk)
  (let ((value
	 (call-with-current-continuation
	  (lambda (throw)
	    (call-with-output-file filename
	      (lambda (out)
		(bind-condition-handler (list condition-type:serious-condition
					      condition-type:simple-condition)
		    (named-lambda (log-condition-handler condition)
		      (write-condition-report condition out)
		      (newline out)
		      (close-port out)
		      (throw condition))
		  (lambda ()
		    (parameterize ((current-output-port out))
		      (thunk))))))
	    #f))))
    (set-file-modes! filename #o444)
    (if (condition? value)
	(error value))))

(define (log-timestamp)
  (log "# "(universal-time->local-time-string (get-universal-time))"\n"))

(define (log . strings)
  (for-each write-string strings))

(define (run . strings)
  (let ((cmdln (string* strings)))
    (log cmdln"\n")
    (shell* cmdln)))

(define (run-noerror . strings)
  (let ((cmdln (string* strings)))
    (log cmdln"\n")
    (shell*-noerror cmdln)))

(define (shell-lines . strings)
  (call-with-input-string (shell-output (string* strings)) read-lines))

(define (file-lines . strings)
  (call-with-input-file (string* strings) read-lines))

(define (read-lines port)
  (let loop ((lines '()))
    (let ((line (read-line port)))
      (if (eof-object? line)
	  (reverse! lines)
	  (loop (cons line lines))))))

(define (shell-output . strings)
  (let ((cmdln (string* strings)))
    (call-with-output-string
     (lambda (port)
       (let ((status (shell* cmdln 'output port)))
	 (if (not (zero? status))
	     (error "Non-zero exit status:" cmdln)))))))

(define (shell . strings)
  (shell* (string* strings)))

(load-option 'synchronous-subprocess)

(define (shell* cmdln . options)
  (let ((status (apply run-shell-command cmdln
		       'environment scheme-subprocess-environment
		       'redisplay-hook flush-output-port
		       options)))
    (if (not (zero? status))
	(error "Shell command failed:" status cmdln))
    status))

(define (shell*-noerror cmdln . options)
  (apply run-shell-command cmdln
	 'environment scheme-subprocess-environment
	 'redisplay-hook flush-output-port
	 options))

(define (with-subprocess-environment-variable name value thunk)
  (let* ((outside scheme-subprocess-environment)
	 (inside (setenv! name value outside)))
    (dynamic-wind
     (lambda ()
       (set! scheme-subprocess-environment inside))
     thunk
     (lambda ()
       (set! scheme-subprocess-environment outside)))))

(define (setenv! name value env)
  (let* ((prefix (string name"="))
	 (i (let ((end (vector-length env)))
	      (let loop ((i 0))
		(and (fix:< i end)
		     (if (string-prefix? prefix (vector-ref env i))
			 i
			 (loop (fix:1+ i))))))))
    (if i
	(let ((new (vector-copy env)))
	  (vector-set! new i (string prefix value))
	  new)
	(let* ((new-i (vector-length env))
	       (new (vector-grow env (fix:1+ new-i))))
	  (vector-set! new new-i (string prefix value))
	  new))))

(define (set-subprocess-environment-variable! name value)
  (set! scheme-subprocess-environment
	(setenv! name value scheme-subprocess-environment)))

(define (delete-subprocess-environment-variable! name)
  (set! scheme-subprocess-environment
	(let ((prefix (string name"=")))
	  (list->vector
	   (filter (lambda (elt)
		     (not (string-prefix? prefix elt)))
		   (vector->list scheme-subprocess-environment))))))

(let ((len (vector-length scheme-subprocess-environment)))
  (let loop ((i 0))
    (if (fix:< i len)
	(let ((setting (vector-ref scheme-subprocess-environment i)))
	  (if (string-prefix? "PATH=" setting)
	      (vector-set!
	       scheme-subprocess-environment i
	       (string "PATH="
		       (decorated-string-append
			"" ":" ""
			(filter (lambda (dir)
				  (not (string-prefix? "/usr/local" dir)))
				(burst-string
				 (string-tail setting 5) #\: #f)))))
	      (loop (fix:1+ i)))))))