#| -*-Scheme-*-

Copyright (C) 2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a devops plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Developer Operations

;;; See devops.texi for complete details.

(define (main)
  (let* ((args (command-line-arguments))
	 (arg1 (and (pair? args) (car args))))
    (cond ((equal? arg1 "status") (apply devops:status (cdr args)))
	  ((equal? arg1 "release") (apply devops:release (cdr args)))
	  ((equal? arg1 "build") (apply devops:build (cdr args)))
	  ((equal? arg1 "make") (apply devops:make (cdr args)))
	  (else (error "Unknown developer operation:" args)))))

(define (devops:status #!optional name)
  (load "devops/config.scm" '(devops))
  (cond ((default-object? name)
	 (status))
	((string=? "core" name)
	 (core-status (get-core-version) (dirt)))
	(else
	 (plugin-status (->plugin name) (dirt)))))

(define (dirt)
  (shell-lines "git status --porcelain --untracked-files=no"))

(define (status)
  (let ((dirt (dirt))
	(version (get-core-version)))
    (if version
	(core-status version dirt))
    (for-each (lambda (p) (plugin-status p dirt)) plugin-list)))

(define (core-status version dirt)
  (let ((lint (core-lint version (or (get-core-changes) '()) dirt)))
    (if (not (null? lint))
	(begin
	  (log "\n# core "(version-string version)"\n")
	  (write-lint lint)))))

(define (plugin-status plugin dirt)
  (let* ((version (plugin-version plugin))
	 (changes (plugin-changes plugin))
	 (lint (plugin-lint plugin version changes dirt)))
    (if (not (null? lint))
	(begin
	  (log "\n# "(plugin-name plugin)" "(version-string version)"\n")
	  (write-lint lint)))))

(define (write-lint lint)
  (for-each (lambda (line)
	      (write-string line)
	      (newline))
	    lint))

(define core-lint-hook #f)
(define plugin-lint-hook #f)

(define (core-lint version changes dirt)
  (let ((dirt (core-dirt dirt)))
    (append
     (debian-version-lint version "src/runtime/version.scm"
			  (debian-version ".") "debian/changelog")
     (released-version-lint version (released-version (project-name))
			    changes "src/runtime/version.scm")
     (if core-lint-hook
	 (core-lint-hook version changes dirt)
	 '())
     (dirt-lint dirt)
     (changes-lint changes))))

(define (plugin-lint plugin version changes dirt)
  (let ((pdirt (plugin-dirt plugin dirt))
	(pkg (plugin-package plugin))
	(dir (plugin-directory plugin)))
    (append
     (debian-version-lint version (string dir"/configure.ac")
			  (debian-version dir) (string dir"/debian/changelog"))
     (news-version-lint plugin version)
     (subsystem-version-lint plugin version)
     (released-version-lint version (released-version pkg) changes
			    (string (plugin-directory plugin)"/configure.ac"))
     (if plugin-lint-hook
	 (plugin-lint-hook plugin version changes dirt)
	 '())
     (dirt-lint pdirt)
     (changes-lint changes))))

(define debian-changelog-version-pattern
  (compile-regsexp '(seq (* (any-char))
			 #\space #\(
			 (group version (+ (char-not-in #\))))
			 #\))))

(define (debian-version dir)
  (let* ((changelog (string dir"/debian/changelog"))
	 (line (and (file-exists? changelog)
		    (read-first-line changelog)))
	 (match (and line
		     (regsexp-match-string debian-changelog-version-pattern
					   line))))
    (and match
	 (->version (match-extract match 'version)))))

(define (debian-version-lint version filename dversion dfilename)
  (append
   (if (not version)
       (list (string filename": package version not found."))
       '())
   (if (not dversion)
       (list (string dfilename": Debian version not found."))
       '())
   (if (and version dversion
	    (not (version=? version dversion)))
       (list (string dfilename": Debian version"
		     " ("(version-string dversion)")"
		     " does not match."))
       '())))

(define (news-version-lint plugin version)
  (let ((nvers (read-news-version plugin)))
    (if nvers
	(if (string=? nvers (version-string version))
	    '()
	    (list (string (plugin-directory plugin)"/NEWS:"
			  " version ("nvers") does not match.")))
	(list (string (plugin-directory plugin)"/NEWS:"
		      " version not found.")))))

(define (subsystem-version-lint plugin version)
  (let ((svers (read-subsystem-version plugin)))
    (if svers
	(if (version=? svers version)
	    '()
	    (list (string (plugin-directory plugin)"/make.scm:"
			  " subsystem version "svers
			  " does not match.")))
	(list (string (plugin-directory plugin)"/make.scm:"
		      " subsystem version not found.")))))

(define (released-version-lint version released changes filename)
  (cond ((eq? #f released)
	 (list "First release!"))
	((and (pair? changes)
	      (not (version>? version released)))
	 (list (string filename": version is out-of-date.")))
	(else
	 '())))

(define (dirt-lint dirt)
  (if (pair? dirt)
      (cons "Uncommitted files:" dirt)
      '()))

(define (changes-lint changes)
  (if (pair? changes)
      (cons "Unreleased commits:" changes)
      '()))

(define (core-dirt dirt)
  (filter (lambda (line)
	    (not (any (lambda (plugin) (plugin-dirt? plugin line))
		      plugin-list)))
	  dirt))

(define (plugin-dirt plugin dirt)
  (filter (lambda (line)
	    (plugin-dirt? plugin line))
	  dirt))

(define (plugin-dirt? plugin line)
  (let* ((dir (plugin-directory plugin))
	 (dir-len (string-length dir))
	 (dir-end (fix:+ 3 dir-len)))
    (and (fix:> (string-length line) dir-end)
	 (string=? dir (string-slice line 3 dir-end))
	 (char=? #\/ (string-ref line dir-end)))))

(define core-version-pattern
  (compile-regsexp '(seq (* (any-char))
			 "ubsystem-identification! \"Release\""
			 " '("
			 (group version (+ (char-in #\space numeric)))
			 ")")))

(define (get-core-version)
  (and (file-exists? "src/runtime/version.scm")
       (call-with-input-file "src/runtime/version.scm"
	 (lambda (in)
	   (let loop ()
	     (let* ((line (read-line in))
		    (match (and (string? line)
				(regsexp-match-string core-version-pattern
						      line))))
	       (if match
		   (map (lambda (s) (string->number s 10))
			(burst-string (match-extract match 'version)
				      #\space #t))
		   (if (eof-object? line)
		       (error "Could not find core version.")
		       (loop)))))))))

(define (make-news-pattern fullname)
  (compile-regsexp `(seq ,fullname
			 #\space
			 (group version (+ (char-in #\. numeric)))
			 " - "
			 (group author (+ (char-not-in #\,)))
			 ", "
			 (group year (+ (char-in numeric)))
			 "-"
			 (group month (+ (char-in numeric)))
			 "-"
			 (group day (+ (char-in numeric)))
			 (string-end))))

(define (read-news-version plugin)
  (let* ((fullname (string (project-name)"-"(plugin-name plugin)))
	 (file (string (plugin-directory plugin)"/NEWS")))
    (and (file-exists? file)
	 (let ((pattern (make-news-pattern fullname)))
	   (call-with-input-file file
	     (lambda (in)
	       (let loop ()
		 (let ((line (read-line in)))
		   (if (eof-object? line)
		       #f
		       (let ((match (regsexp-match-string pattern line)))
			 (if match
			     (match-extract match 'version)
			     (loop))))))))))))

(define subsystem-pattern
  (compile-regsexp
   '(seq "(add-subsystem-identification!"
	 " \""(+ (char-not-in #\"))"\""
	 " '"(group version (seq "("
				 (+ (char-in #\space numeric))
				 ")"))
	 ")")))

(define (read-subsystem-version plugin)
  (let ((file (find-plugin-make.scm plugin)))
    (and file
	 (call-with-input-file file
	   (lambda (in)
	     (let loop ()
	       (let ((line (read-line in)))
		 (if (eof-object? line)
		     #f
		     (let ((match
			    (regsexp-match-string subsystem-pattern line)))
		       (if match
			   (call-with-input-string
			       (match-extract match 'version)
			     read)
			   (loop)))))))))))

(define (find-plugin-make.scm plugin)
  (or (let ((file (string (plugin-directory plugin)"/make.scm")))
	(and (file-exists? file)
	     file))
      (let ((file (string (plugin-directory plugin)"/mit-make.scm")))
	(and (file-exists? file)
	     file))))

(define (released-version name)
  (let ((tags (sorted-tags name)))
    (and (pair? tags)
	 (caar tags))))

(define (get-core-changes)
  (let ((start-hash (released-hash (project-name))))
    (and start-hash
	 (append-map!
	  (lambda (hash)
	    (let* ((lines
		    (shell-lines "git log --oneline --name-status -1 "hash))
		   (files
		    (filter (let ((excludes (plugin-dir-prefixes)))
			      (lambda (line)
				(let* ((i (string-find-next-char line #\tab))
				       (name (string-slice line (fix:1+ i))))
				  (not (any (lambda (exclude)
					      (string-prefix? exclude name))
					    excludes)))))
			    (cdr lines))))
	      (if (null? files)
		  '()
		  (cons (car lines) files))))
	  (shell-lines "git log --format=%H "start-hash"..HEAD")))))

#;(define (core-changed-files)
  (let ((hash (released-hash (project-name))))
    (and hash
	 (let ((excluded-dirs (map plugin-directory plugin-list)))
	   (filter
	    (lambda (filename)
	      (not (any (lambda (excluded-dir)
			  (string-prefix? excluded-dir filename))
			excluded-dirs)))
	    (shell-lines "git diff --name-only "hash))))))

(define (plugin-dir-prefixes)
  (map (lambda (p) (string-append (plugin-directory p) "/"))
       plugin-list))

(define (released-hash name)
  (let* ((tags (sorted-tags name))
	 (last-tag (and (pair? tags)
			(cdar tags))))
    (and last-tag
	 (car (shell-lines "git log --format=%H -1 "last-tag)))))

(define plugin-version-pattern
  (compile-regsexp '(seq (* (char-in whitespace))
			 #\[
			 (group version (+ (char-in #\. numeric)))
			 #\])))

(define (plugin-version plugin)
  (call-with-input-file (string (plugin-directory plugin)"/configure.ac")
    (lambda (in)
      (let loop ()
	(let ((line (read-line in)))
	  (if (and (string? line)
		   (string-prefix? "AC_INIT" line))
	      (let* ((line (read-line in))
		     (match (and (string? line)
				 (regsexp-match-string plugin-version-pattern
						       line))))
		(if match
		    (->version (match-extract match 'version))
		    (error "No plugin version:" (plugin-name plugin))))
	      (if (eof-object? line)
		  (error "No AC_INIT:" (plugin-name plugin))
		  (loop))))))))

(define (plugin-changes plugin)
  (let ((hash (released-hash (plugin-package plugin))))
    (and hash
	 (shell-lines "git log --no-merges --oneline --name-status "hash".."
		      " -- "(plugin-directory plugin)"/"))))

#;(define (plugin-changed-files plugin)
  (let ((hash (released-hash (plugin-package plugin))))
    (and hash
	 (let ((dir (plugin-directory plugin)))
	   (filter (lambda (filename)
		     (string-prefix? dir filename))
		   (shell-lines "git diff --name-only "hash))))))

;;;; Release

(define tag-create-options "")

(define (tag-options . options)
  (guarantee-list-of string? options 'tag-options)
  (if (not (null? options))
      (set! tag-create-options ((string-joiner* 'infix " ") options)))
  tag-create-options)

(define (devops:release #!optional plugin)
  (load "devops/config.scm" '(devops))
  (let ((dirt (shell-lines "git status --porcelain --untracked-files=no")))
    (cond ((default-object? plugin)
	   (release-core (get-core-version) (get-core-changes) dirt #f))
	  ((or (equal? "snapshot" plugin)
	       (eq? 'snapshot plugin))
	   (snapshot-core dirt)
	   (for-each (lambda (p) (snapshot-plugin p dirt))
		     plugin-list))
	  ((or (string? plugin) (symbol? plugin))
	   (let ((p (->plugin plugin)))
	     (release-plugin p (plugin-version p) (plugin-changes p) dirt #f)))
	  (else
	   (error "Plugin must be a string or symbol.")))))

(define (release-core version changes dirt snap?)
  (if (and (null? changes) (not snap?))
      (error "No unreleased commits."))
  (let* ((project (project-name))
	 (vers (version-string version))
	 (pkgvers (string project"-"vers))
	 (out (current-output-port))
	 (runl (lambda strings (runl* out strings)))
	 (logfile (string "devops/"pkgvers"-src.log")))
    (log "# "pkgvers":\n")
    (write-lint (core-lint version '() dirt))
    (with-output-log
     logfile
     (lambda ()
       (runl "mkdir devops/"pkgvers)
       (if snap?
	   (runl "git archive --prefix="project"/ HEAD"
		 " | ( cd devops/"pkgvers" && tar xf - )")
	   (begin
	     (runl "git tag "tag-create-options" -m \"\" "pkgvers)
	     (runl "git archive --prefix="project"/ "pkgvers
		   " | ( cd devops/"pkgvers" && tar xf - )")))
       (runl "cd devops/"pkgvers" && "project"/dist/make-src-files standard")
       (runl "mv devops/"pkgvers"/"pkgvers".tar.gz devops/")
       (runl "rm -rf devops/"pkgvers)
       (runl "cd devops/ && tar xzf "pkgvers".tar.gz")
       (runl "cd devops/ && dpkg-source --build "pkgvers)
       (runl "chmod 444 devops/"project"_"vers".dsc")
       (runl "chmod 444 devops/"project"_"vers".tar.xz")
       (runl "rm -rf devops/"pkgvers)))
    (delete-file logfile)))

(define (runl* out strings)
  (let ((cmdln (string* strings)))
    (write-string cmdln out)(newline out)
    (write-string cmdln)(newline)
    (shell* cmdln)))

(define (release-plugin plugin version changes dirt snap?)
  (if (and (null? changes) (not snap?))
      (error "No unreleased commits."))
  (let* ((vers (version-string version))
	 (pkg (plugin-package plugin))
	 (pkgvers (string pkg"-"vers))
	 (out (current-output-port))
	 (runl (lambda strings (runl* out strings)))
	 (logfile (string "devops/"pkgvers"-src.log"))
	 (dir (plugin-directory plugin)))
    (log "# "pkgvers":\n")
    (write-lint (plugin-lint plugin version '() dirt))
    (with-output-log
     logfile
     (lambda ()
       (if snap?
	   (runl "git archive --prefix="pkgvers"/ HEAD -- "dir
		 " | ( cd devops && tar xf - )")
	   (begin
	     (runl "git tag "tag-create-options" -m \"\" "pkgvers)
	     (runl "git archive --prefix="pkgvers"/ "pkgvers" -- "dir
		   " | ( cd devops && tar xf - )")))
       (runl "cd devops/"pkgvers"/"dir" && ./autogen.sh")
       (runl "cd devops/"pkgvers"/"dir" && ./configure")
       (runl "cd devops/"pkgvers"/"dir" && make dist")
       (runl "mv devops/"pkgvers"/"dir"/"pkgvers".tar.gz devops/")
       (runl "chmod 444 devops/"pkgvers".tar.gz")
       (runl "rm -rf devops/"pkgvers)
       (runl "cd devops/ && tar xzf "pkgvers".tar.gz")
       (runl "cd devops/ && dpkg-source --build "pkgvers)
       (runl "chmod 444 devops/"pkg"_"vers".dsc")
       (runl "chmod 444 devops/"pkg"_"vers".tar.xz")
       (runl "rm -rf devops/"pkgvers)))
    (delete-file logfile)))

(define (snapshot-core version dirt)
  (let* ((changes (or (get-core-changes) '()))
	 (released (released-version (project-name)))
	 (vers (version-string version))
	 (source-filename (string "devops/"(project-name)"-"vers".tar.gz")))
    (define (found) (log "# "source-filename":\nAlready done.\n"))
    (cond ((and (null? changes)
		(version=? version released))
	   (if (file-exists? source-filename)
	       (found)
	       (release-core version changes dirt #t)))
	  ((and (pair? changes)
		(version>? version released))
	   ;; Clobber!
	   (release-core version changes dirt #t))
	  ((null? changes)
	   (if (file-exists? source-filename)
	       (found)
	       (release-core version changes dirt #t)))
	  (else
	   (error "Version has not incremented:" version)))))

(define (snapshot-plugin plugin dirt)
  (let* ((changes (or (plugin-changes plugin) '()))
	 (version (plugin-version plugin))
	 (released (released-version (plugin-package plugin)))
	 (vers (version-string version))
	 (source-filename (string "devops/"
				  (plugin-package plugin)"-"vers".tar.gz")))
    (define (found) (log "# "source-filename":\nAlready done.\n"))
    (cond ((and (null? changes)
		(version=? version released))
	   (if (file-exists? source-filename)
	       (found)
	       (release-plugin plugin version changes dirt #t)))
	  ((and (pair? changes)
		(version>? version released))
	   ;; Clobber!
	   (release-plugin plugin version changes dirt #t))
	  ((null? changes)
	   (if (file-exists? source-filename)
	       (found)
	       (release-plugin plugin version changes dirt #t)))
	  (else
	   (error "Version has not incremented:" version)))))

;;;; Build

(define (devops:build #!optional hostname)
  (load "devops/config.scm" '(devops))
  (let ((srcs (available-sources "devops"))
	(hosts (hosts)))
    (if (null? hosts)
	(error "No build hosts defined.")
	(if (default-object? hostname)
	    (for-each (lambda (host) (write-host-status host srcs))
		      hosts)
	    (let* ((name
		    (cond ((string? hostname) hostname)
			  ((symbol? hostname) (symbol->string hostname))
			  (else (error "Hostname is not a symbol or string:"
				       hostname))))
		   (host (find (lambda (host)
				 (string=? name (host-name host)))
			       hosts)))
	      (if host
		  (write-host-status host srcs)
		  (error "Build host not found:" hostname)))))))

(define build-dir "devops")

(define reply-wait 10000)

(define (read-reply query i/o)
  (write-line `(write-line ,query) i/o)
  (flush-output-port i/o)
  (let ((object (read-until reply-wait i/o)))
    (if (eq? object 'timeout)
	(error "Timeout awaiting reply:" query))
    object))

(define (write-host-status host srcs)
  (log "# "(host-name host)":\n")
  (call-with-host-i/o
   host
   (lambda (i/o)
     (let ((reply (read-reply '(quote ready) i/o)))
       (if (not (eq? reply 'ready))
	   (error "Build host not ready:" reply)))
     (let ((reply (read-reply `(let ((dirname ,(host-directory host)))
				 (if (and (file-exists? dirname)
					  (file-directory? dirname))
				     'yes
				     'no))
			      i/o)))
       (cond ((eq? reply 'yes)
	      unspecific)
	     ((eq? reply 'no)
	      (error "No build directory:" (host-directory host)))
	     (else (error "Unexpected reply:" reply))))
     (let ((files (read-reply `(directory-file-names ,(host-directory host) #f)
			      i/o))
	   (proj (project-name))
	   (sarch (host-scheme-architecture host))
	   (darch (host-debian-architecture host)))
       (if (not (list-of-type? files string?))
	   (error "Unexpected filename list:" files))
       (write-lock-report host i/o files)
       (for-each
	 (lambda (src)
	   (let ((name (car src))
		 (vers (cdr src)))
	     (if (not (member (string name"-"vers".tar.gz") files))
		 (run "scp -p devops/"name"-"vers".tar.gz"
		      " "(host-login/dir host))
		 (if (string=? proj name)
		     (write-pkg-status name vers sarch files host i/o)))

	     (if (host-build-ubuntu? host)
		 (let ((dsc (string name"_"vers".dsc"))
		       (tar (string name"_"vers".tar.xz")))
		   (if (or (not (member dsc files))
			   (not (member tar files)))
		       (begin
			 (if (not (member dsc files))
			     (run "scp -p devops/"dsc" "(host-login/dir host)))
			 (if (not (member tar files))
			     (run "scp -p devops/"tar" "(host-login/dir host))))
		       (write-ubuntu-status name vers darch files host i/o))))))
	 srcs))
     (start-builds host i/o))))

(define (write-lock-report host i/o files)
  (if (member "lock" files)
      (let* ((lockfile (string (host-directory host)"/lock"))
	     (start-time
	      (read-reply `(file-time->local-time-string
			    (file-modification-time ,lockfile))
			  i/o)))
	(log "Daemon started "start-time"\n"))
      (log "No daemon running.\n")))

(define (write-pkg-status name vers sarch files host i/o)
  (let ((logfile (string name"-"vers"-"sarch"-pkg.log")))
    (if (not (member logfile files))
	(let ((bin  (string name"-"vers"-"sarch".tar.gz"))
	      (elogfile (string name"-"vers"-"sarch"-pkg-error.log")))
	  (log bin"\n")
	  (if (member elogfile files)
	      (write-minutes-stalled
	       (string (host-directory host)"/"elogfile) i/o))))))

(define (write-ubuntu-status name vers darch files host i/o)
  (let ((logfile (string name"-"vers"-"darch"-dpkg.log")))
    (if (not (member logfile files))
	(let ((bin (string name"_"vers"_"darch".deb"))
	      (elogfile (string name"_"vers"_"darch"-dpkg-error.log")))
	  (log bin"\n")
	  (if (member elogfile files)
	      (write-minutes-stalled
	       (string (host-directory host)"/"elogfile) i/o))))))

(define (write-minutes-stalled elogpath i/o)
  (let* ((sec (read-reply `(- (get-universal-time)
			      (file-time->universal-time
			       (file-modification-time ,elogpath)))
			  i/o))
	 (min (quotient sec 60)))
    (cond ((> min 60)
	   (log "    No progress in over an hour!\n"))
	  ((> min 10)
	   (log "    No progress in "(number->string min 10)" minutes.\n"))
	  (else
	   (log "    Working...\n")))))

(define (start-builds host i/o)
  (write-file (let ((repo "src/devops/build.scm"))
		(if (file-exists? repo)
		    repo
		    (let ((sys (system-library-pathname "devops/build.scm" #f)))
		      (if (file-exists? sys)
			  sys
			  (error "Could not find build script.")))))
	      i/o)

  (verify-host-architecture host i/o)

  (write-line `(begin
		 (set! project-name ,(project-name))
		 (set! build-dir ,(host-directory host))
		 (set! build-scheme-architecture
		       ',(host-scheme-architecture host))
		 (set! build-debian-architecture
		       ,(host-debian-architecture host))
		 (set! build-ubuntu?
		       ,(host-build-ubuntu? host))
		 (build))
	      i/o)
  (flush-output-port i/o)
  (let ((reply (read-reply '(quote ok) i/o)))
    (if (not (eq? reply 'ok))
	(error "Build host not OK.")))
  (close-input-port i/o)
  (close-output-port i/o))

(define (verify-host-architecture host i/o)
  (if (host-build-ubuntu? host)
      (begin
	(let ((darch (read-reply '(read-debian-architecture) i/o)))
	  (if (not (string? darch))
	      (error "No Debian architecture."))
	  (if (not (string=? darch (host-debian-architecture host)))
	      (error "Wrong Debian architecture.")))
	(let ((str (read-reply '(if (ubuntu?) "yes" "no") i/o)))
	  (if (not (string? str))
	      (error "No Ubuntu-ness."))
	  (let ((ubu? (string=? "yes" str)))
	    (if (not (eq? ubu? (host-ubuntu? host)))
		(error "Wrong Ubuntu-ness.")))))))

(define (call-with-host-i/o host receiver)
  (call-with-current-continuation
   (lambda (punt)
     (bind-condition-handler (list condition-type:serious-condition
				   condition-type:simple-condition)
	 (named-lambda (host-condition-handler condition)
	   (write-condition-report condition (current-output-port))
	   (newline)
	   (punt unspecific))
       (lambda ()
	 (let ((proc #f))
	   (dynamic-wind
	    (lambda ()
	      unspecific)
	    (lambda ()
	      (set! proc (start-pipe-subprocess
			  (os/find-program "ssh" #f)
			  (vector "ssh" (host-login host)
				  "mit-scheme" "--batch-mode")
			  #f))
	      (receiver (subprocess-i/o-port proc)))
	    (lambda ()
	      (if (and proc (memq (subprocess-status proc) '(running stopped)))
		  (ignore-errors (lambda () (subprocess-kill proc))))))))))))

(define (read-lines-until line usec in)
  (do-until
   (lambda ()
     (let loop ((lines '()))
       (let ((line* (read-line in)))
	 (if (eof-object? line*)
	     (reverse! lines)
	     (if (string=? line line*)
		 (reverse! (cons line* lines))
		 (loop (cons line* lines)))))))
   usec
   (lambda () #f)))

(define (read-until usec in)
  (do-until (lambda () (read in))
	    usec
	    (lambda () 'timeout)))

(define (do-until thunk usec timeout)
  (call-with-current-continuation
   (lambda (return)
     (let* ((record
	     (register-timer-event usec
				   (named-lambda (timeout-do-until)
				     (return (timeout)))))
	    (value (thunk)))
       (deregister-timer-event record)
       value))))

;;;; Make

(define (devops:make target)
  (if (not (member target '(#f "svm" "C" "x86-64" "i386")))
      (error "Unknown build target:" target))
  (load-make-config)
  (exit
   (call-with-current-continuation	;throw here to unwind all
    (lambda (abort-job)
      (bind-condition-handler (list condition-type:error)
	  (named-lambda (job-error-handler condition)
	    (fresh-line)
	    (log "# "(emacs-friendly-timestamp)"\n")
	    (write-condition-report condition (current-output-port))
	    (newline)
	    (flush-output-port)
	    (abort-job #f))
	(lambda ()
	  (with-^G-interrupt-handler
	   (named-lambda (job-^G-interrupt-handler)
	     (fresh-line)
	     (log "# "(emacs-friendly-timestamp)"\n# job aborted\n")
	     (flush-output-port)
	     (abort-job #f))
	   (lambda ()
	     (devops:make* target)
	     #t))))))))

(define (devops:make* target)
  (let* ((prefix (or (get-environment-variable "PWD")
		     (error "PWD not set.")))
	 (host-exe (or (get-environment-variable "MIT_SCHEME_EXE")
		       "mit-scheme"))
	 (a.v (read-host-arch.version host-exe))
	 (host-arch (symbol->string (car a.v)))
	 (host-version (cdr a.v))
	 (master? (cond ((file-directory? ".git")
			 (not (file-directory? "debian")))
			((file-directory? "../.git")
			 (not (file-directory? "../debian")))
			(else (error "Not a git working directory:"
				     (working-directory-pathname))))))

    ;;(set-environment-variable! "LD_LIBRARY_PATH" "/usr/local/lib")
    (set-subprocess-environment-variable! "FAST" "please")
    (delete-subprocess-environment-variable! "DISPLAY")

    (log "# Host: "host-arch" "(version-string host-version)" "host-exe"\n")
    (log "# Target: "target"\n")
    (log "# Config: "make-config"\n")

    (cond ((file-directory? ".git")
	   (make-install-doc prefix))
	  ((file-directory? "../.git")
	   (lndir "../doc")
	   (make-install-doc prefix))
	  (else
	   (error "Not a git working directory:"
		  (working-directory-pathname))))

    (cond ((file-directory? ".git"))
	  ((file-directory? "../.git")
	   (lndir "../src"))
	  (else
	   (error "Unexpected source directory:"(working-directory-pathname))))
    (if (not (file-exists? "src/configure"))
	(trun "cd src/ && ./Setup.sh"))
    (if (not (file-exists? "src/Makefile"))
	(trun "cd src/ && ./configure --prefix="prefix
	      (if (not master?) " --disable-default-plugins" "")
	      (if (not (string-null? make-config)) " " "")
	      make-config))
    (trun "cd src/ && make tags")
    (trun "cd src/ && make all")
    (lndir "../tests")
    (trun-noerror "cd src/ && make check")
    (trun "cd src/ && umask 022 && make install")

    (let ((plugin-errors
	   (with-EXE-environment-variables
	    master? (string prefix"/bin/"(project-name))
	    (lambda ()
	      (reduce + 0
		      (map (lambda (plugin)
			     (make-install-plugin plugin prefix master?))
			   plugin-list))))))
      (if (not (zero? plugin-errors))
	  (error "Plugins failed to build:" plugin-errors)))))

(define (with-EXE-environment-variables master? exe thunk)
  (with-subprocess-environment-variable
   "MIT_SCHEME_EXE" exe
   (lambda ()
     (if master?
	 (with-subprocess-environment-variable
	  "MIT_SCHEME_TOOLCHAIN_EXE" exe
	  thunk)
	 (thunk)))))

(define (read-host-arch.version exe)
  (let ((expr (string "(write-line"
		      " (cons microcode-id/compiled-code-type"
		      "       (get-subsystem-version \"Release\")))")))
    (read (open-input-string
	   (car (shell-lines exe " --batch-mode"
			     " --eval '"expr"' --eval '(%exit)'"))))))

(define (load-make-config)
  (cond ((file-exists? "devops-config.scm")
	 (load "devops-config.scm" '(devops)))
	((file-exists? "../devops-config.scm")
	 (load "../devops-config.scm" '(devops)))
	((let* ((git-root (find-git-root))
		(file (string git-root"/devops-config.scm")))
	   (and (file-exists? file)
		(begin (load file '(devops))
		       #t))))
	(else
	 (error "No devops test configuration found."))))

(define make-config "")
(define (make-configuration . args)
  (guarantee-list-of string? args 'make-configuration)
  (if (not (null? args))
      (set! make-config ((string-joiner* 'infix " ") args)))
  make-config)

(define (lndir dest)
  (let ((destdir (dirname dest))
	(name (filename dest)))
    (lndir* name destdir "" (drop-slash destdir))))

(define (lndir* name dstdir srcdir reldir)
  (let ((dst (string dstdir name))
	(src (string srcdir name)))
    (let ((dst-atts (file-attributes-direct dst))
	  (src-atts (file-attributes-direct src)))
      (if (eq? #t (file-attributes/type dst-atts))
	  (if (not (or (eq? #f src-atts)
		       (eq? #t (file-attributes/type src-atts))))
	      (warn "not a directory:" src)
	      (let ((dstdir* (string dst"/"))
		    (srcdir* (string src "/"))
		    (reldir* (string "../"reldir"/"name)))
		(if (eq? #f src-atts)
		    (make-directory src))
		(for-each
		  (lambda (name)
		    (cond ((string=? "." name))
			  ((string=? ".." name))
			  ((string-suffix? "~" name))
			  (else (lndir* name dstdir* srcdir* reldir*))))
		  (directory-file-names dst))))
	  (let ((reldirname (string reldir"/"name)))
	    (if (eq? #f src-atts)
		(soft-link-file reldirname src)
		(let ((src-type (file-attributes/type src-atts)))
		  (cond ((not (string? src-type))
			 (warn "not a symlink:" src))
			((not (string=? src-type reldirname))
			 (warn "bogus symlink:" src reldirname))))))))))

(define (make-install-doc prefix)
  (if (not (file-exists? "doc/configure"))
      (trun "cd doc/ && autoconf"))
  (if (not (file-exists? "doc/Makefile"))
      (trun "cd doc/ && ./configure --prefix="prefix))
  (trun "cd doc/ && make all")
  (trun "cd doc/ && umask 022 && make install"))

(define default-plugin-names '("edwin" "imail" "x11" "x11-screen"))

(define (make-install-plugin plugin prefix master?)
  (if (and master?
	   (member (plugin-name plugin) default-plugin-names))
      (begin
	(log "# "(plugin-name plugin)" skipped default plugin\n")
	0)
      (let ((result (ignore-errors
		     (lambda () (make-install-plugin* plugin prefix)))))
	(if (condition? result)
	    (begin
	      (log "# "(plugin-name plugin)" failed to build:"
		   " "(condition/report-string result)"\n")
	      1)
	    0))))

(define (make-install-plugin* plugin prefix)
  (let ((dir (plugin-directory plugin)))
    (if (not (file-directory? dir))
	(error "Plugin directory not found:" dir))
    (log "# "dir":\n")
    (if (not (file-exists? (string dir"/configure")))
	(trun "cd "dir"/ && ./autogen.sh"))
    (if (not (file-exists? (string dir"/Makefile")))
	(trun "cd "dir"/ && ./configure --prefix="prefix))
    (trun "cd "dir"/ && make tags")
    (trun "cd "dir"/ && make all")
    (trun "cd "dir"/ && make check")
    (trun "cd "dir"/ && umask 022 && make install")
    (if (find (lambda (line) (string-prefix? "info_TEXINFOS" line))
	      (file-lines dir"/Makefile.am"))
	(begin
	  (trun "cd "dir"/ && umask 022 && make install-html")
	  (trun "cd "dir"/ && umask 022 && make install-pdf")))))

(define (trun . strings)
  ;;(log-timestamp)
  (log "# "(emacs-friendly-timestamp)"\n")
  (apply run strings))

(define (trun-noerror . strings)
  (log "# "(emacs-friendly-timestamp)"\n")
  (let* ((cmdln (string* strings))
	 (status (run-noerror cmdln)))
    (if (not (zero? status))
	(warn "Shell command failed:" status cmdln))))

(define (emacs-friendly-timestamp)
  (let ((dt (local-decoded-time)))
    (define (pad num) (string-pad-left (number->string num) 2 #\0))
    (let ((year (number->string (decoded-time/year dt)))
	  (month (pad (decoded-time/month dt)))
	  (day (pad (decoded-time/day dt)))
	  (hour (pad (decoded-time/hour dt)))
	  (minute (pad (decoded-time/minute dt)))
	  (second (pad (decoded-time/second dt))))
      (string year"-"month"-"day" "hour"."minute"."second))))

(define (with-^G-interrupt-handler inside-handler thunk)
  (let ((env (->environment '(runtime interrupt-handler)))
	(outside-handler))
    (dynamic-wind
     (lambda ()
       (set! outside-handler (environment-lookup env 'hook/^G-interrupt))
       (environment-assign! env 'hook/^G-interrupt inside-handler))
     thunk
     (lambda ()
       (set! inside-handler (environment-lookup env 'hook/^G-interrupt))
       (environment-assign! env 'hook/^G-interrupt outside-handler)))))

;;;; Configuration

(define (project-name #!optional name)
  (if (default-object? name)
      project-name-string
      (begin
	(if (not (string? name))
	    (error "Project name is not a string:" name))
	(set! project-name-string name)
	name)))

(define project-name-string "new-scheme")

(define (plugin-package plugin)
  (string (project-name)"-"(plugin-name plugin)))

(define (plugin name directory)
  (let ((duplicate (find (lambda (p) (string=? name (plugin-name p)))
			 plugin-list)))
    (if duplicate
	(error (string "Plugin "name" already defined."))))
  (set! plugin-list
	(append! plugin-list
		 (list (make-plugin name directory))))
  unspecific)

(define plugin-list '())

(define (->plugin name)
  (let ((n (string name)))
    (or (find (lambda (p) (string=? n (plugin-name p)))
	      plugin-list)
	(error "No such plugin:" name))))

(define-record-type <plugin>
    (make-plugin name directory)
    plugin?
  (name plugin-name)
  (directory plugin-directory))

(define (host name user directory arch os)
  (let ((duplicate (find (lambda (h) (string=? name (host-name h)))
			 host-list)))
    (if duplicate
	(error (string "Host "name" already defined."))))
  (set! host-list
	(append! host-list
		 (list (make-host name user directory arch os))))
  unspecific)

(define (hosts) (list-copy host-list))

(define host-list '())

(define (host-login/dir h)
  (string (host-login h)":"(host-directory h)))

(define (host-login h)
  (let ((n (host-name h))
	(u (host-user h)))
    (if u
	(string u"@"n)
	n)))

(define-record-type <host>
    (make-host name user directory arch os)
    host?
  (name host-name)
  (user host-user)
  (directory host-directory)
  (arch host-scheme-architecture)
  (os host-os))

(define (host-ubuntu? host)
  (os-ubuntu? (host-os host)))

(define (host-build-ubuntu? host)
  (and (host-ubuntu? host)
       (not (string-prefix? "svm" (host-scheme-architecture host)))))

(define (host-debian-architecture host)
  (let ((arch (host-scheme-architecture host)))
    (cond ((string=? "x86-64" arch) "amd64")
	  ((string=? "i386" arch) "i386")
	  ((string=? "svm1-32le" arch) #f)
	  ((string=? "svm1-32be" arch) #f)
	  ((string=? "svm1-64le" arch) #f)
	  ((string=? "svm1-64be" arch) #f)
	  (else (error "Unknown host architecture:" arch)))))

(define (host-ubuntu-codename host)
  (ubuntu-os-codename (host-os host)))

(define (host-ubuntu-version host)
  (ubuntu-os-version (host-os host)))

(define (os-ubuntu? os)
  (string-prefix? "Ubuntu " os))

(define (ubuntu-os-codename os)
  (cond ((string=? "Ubuntu 19.04" os) "disco")
	((string=? "Ubuntu 18.10" os) "cosmic")
	((string=? "Ubuntu 18.04" os) "bionic")
	((string=? "Ubuntu 17.10" os) "artful")
	((string=? "Ubuntu 17.04" os) "zesty")
	((string=? "Ubuntu 16.10" os) "yakkety")
	((string=? "Ubuntu 16.04" os) "xenial")
	(else (error "Unexpected Ubuntu OS:" os))))

(define (ubuntu-os-version os)
  (cond ((string=? "Ubuntu 19.04" os) "19.04")
	((string=? "Ubuntu 18.10" os) "18.10")
	((string=? "Ubuntu 18.04" os) "18.04")
	((string=? "Ubuntu 17.10" os) "17.10")
	((string=? "Ubuntu 17.04" os) "17.04")
	((string=? "Ubuntu 16.10" os) "16.10")
	((string=? "Ubuntu 16.04" os) "16.04")
	(else (error "Unexpected Ubuntu OS:" os))))

;;;; Misc

(define (sorted-tags package-name)
  (sort (let ((pattern (compile-regsexp
			`(seq ,package-name #\-
			      (group version (+ (char-in #\. numeric)))
			      (string-end)))))
	  (append-map!
	   (lambda (line)
	     (let ((match (regsexp-match-string pattern line)))
	       (if match
		   (list (cons (->version (match-extract match 'version))
			       line))
		   '())))
	   (shell-lines "git tag -l '"package-name"-*'")))
	(lambda (a b) (version>? (car a) (car b)))))

(define (->version string)
  (and string
       (map string->number (burst-string string #\. #f))))

(define (version-string version)
  (if (pair? version)
      (decorated-string-append "" "." ""
       (map (lambda (num) (number->string num 10))
	    version))
      (error "Bad version:" version)))

(define version=? equal?)

(define (version<? v1 v2)
  (let loop ((v1 v1) (v2 v2))
    (cond ((null? v2)
	   #f)
	  ((null? v1)
	   #t)
	  ((< (car v1) (car v2))
	   #t)
	  ((> (car v1) (car v2))
	   #f)
	  (else
	   (loop (cdr v1) (cdr v2))))))

(define (version>? v1 v2)
  (version<? v2 v1))

(define (match-extract match key)
  (let ((entry (assq key (cddr match))))
    (if entry
	(cdr entry)
	(error "Match key not found:" key match))))

(define (read-first-line filename)
  (call-with-input-file filename
    (lambda (in)
      (let ((line (read-line in)))
	(and (string? line)
	     line)))))

(define (write-file filename out)
  (call-with-input-file filename
    (lambda (in)
      (let loop ()
	(let ((line (read-line in)))
	  (if (not (eof-object? line))
	      (begin
		(write-string line out)
		(newline out)
		(loop))))))))

(define (find-git-root)
  (let ((pwd (drop-slash (or (get-environment-variable "PWD")
			     (error "PWD not set.")))))
    (let loop ((parent (drop-slash (dirname pwd))))
      (if (string-null? parent)
	  (error "No git root found:" pwd)
	  (if (file-directory? (string parent"/.git"))
	      parent
	      (loop (drop-slash (dirname pwd))))))))

(define (final-slash? string)
  (let ((len-1 (fix:-1+ (string-length string))))
    (and (fix:>= len-1 0)
	 (char=? #\/ (string-ref string len-1)))))

(define (drop-slash string)
  (let ((len-1 (fix:-1+ (string-length string))))
    (if (and (fix:>= len-1 0)
	     (char=? #\/ (string-ref string len-1)))
	(string-slice string 0 len-1)
	string)))

(define (add-slash string)
  (if (not (final-slash? string))
      (string-append string "/")))

(define (filename filename)
  (let ((i (string-find-previous-char filename #\/)))
    (if (not i)
	filename
	(string-slice filename (fix:1+ i)))))

(define (dirname filename)
  (let ((i (string-find-previous-char filename #\/)))
    (if (not i)
	""
	(string-slice filename 0 (fix:1+ i)))))

(define (basename filename)
  (let* ((start (let ((i (string-find-previous-char filename #\/)))
		  (if (not i)
		      0
		      (fix:1+ i))))
	 (end (string-find-next-char filename #\. start)))
    (if (not end)
	(if (fix:zero? start)
	    filename
	    (string-slice filename start))
	(string-slice filename start end))))